#!/bin/env python
# Generate the xfst code which will in turn generate the
# (perhaps partial) paradigm of a word, in XML format.
#
# Arguments:
#      -s <stem>   #(obligatory) a string which gives the citation form of the word
#                  # Used for display purposes (to tell what word this is the paradigm of).
#      -r <regex>  #(obligatory) A regex (string) which selects the stem of the word,
#                  # in whatever form the xfst grammar expects for the UPPER side
#                  # (which may be a citation form, gloss, UID, etc.).
#                  # This is often an expr like
#                  #    ${have/V} .o. Lex
#                  # It does NOT need to have the final semicolon
#      -f <fname>  #(obligatory) A file containing a list containing
#                  # as its sublists the feature structures representing
#                  # the rows and columns of the paradigm that we want
#                  # to generate (see below for details and examples)
#      -e <encoding> #(optional) Encoding for the language; defaults
#                  # to ISO-8859-1.  (Used in XML heading.)
#      -o <fname>  #(optional) An xfst file to output (defaults to stdout)
#      -h          #Instead of the above, this will output a help msg,
#                  # then exit.
#
# No attempt is made here to rule out impossible combinations of feature
# structures, on the grounds that the purpose of this program is to test
# the xfst grammar, and it is the job of the latter (or some auxilliary program)
# to rule out illicit combinations.
#
# -----------Input format:
# The list of feature structures should contain the upper-side names of the affixes,
# in space-delimited form, grouped into lines; the features of each line represents
# a single set of alternatives, usually the alternatives for a slot.  (A line may
# be continued by ending it with the backslash char "\".)
# A '^' as the first symbol in a line means that no wordforms containing
# any of the other features in the line will be output (useful when outputting
# a sub-paradigm and a slot is, in essence, optional).
# A '0' (zero) at any place in the line means that in addition to outputting cells
# with the forms that contain the other features of that line, a cell will be output
# that contains none of the other features in the line (useful when a slot is
# optional, and it is desired to output both the forms containing the affixes
# of that slot, as well as the forms not containing any of those affixes).
#
# Comments may follow a '#' symbol.
# Example:
#     1SgSubj- 2SgSubj- 3SgSubj- 1PlSubj- 2PlSubj- 3PlSubj-
#     1SgObj-  2SgObj-  3SgObj-  1PlObj-  2PlObj-  3PlObj- 0
#     Drxnl- 0  #The '0' accounts for the optionality of the Drxnl- suffix
#     PresTns-  PastTense-  FutTense-
#
# -----------Output format:
# The immediate output consist of a series of xfst commands to generate the
# surface forms (if any) corresponding to each combination of feature structures,
# together with the XML tags.  If run through xfst (with a lexicon loaded and
# assigned to the var 'Lex'), this will give the following XML:
#    <?xml version="1.0" encoding="ISO-8859-1"?>
#    <!--Generated by GenParadigm.py from Sample.features-->
#    <?xml-stylesheet type="text/css" href="Paradigm.xsl"?>
#    <Paradigm stem="foo">
#       <Cell>
#           <Features> ... </Features>
#           <Wordforms>
#   (wordforms, one per line)
#           </Wordforms>
#       </Cell>
#    </Paradigm>
# It would be nice to put the wordforms that xfst outputs inside individual XML
# tags.  Unfortunately, we can't control xfst's output in that way.  However,
# the xfst 'echo' cmd produces lines with a leading space, whereas the generated
# words start in column 1, and contain no blanks.  So putting the generated words
# inside XML tags can be done by a 'sed' cmd like the ff. operating on the output:
#      	sed -e 's%^[^<> ][^<> ]*$$%       <Wordform form="&"/>%' \
# (The regex "[^<> ]" means anything but a '<', '>', or a space; the repetition
# with Kleene star is necessary because 'sed' lacks Kleene plus.)


import sys
import getopt
import string, re
import copy
import time     #To timestamp output file

#------------------------------functions-------------------------------

def GetCmdParameters():
    sUsageMsg = "Usage: GenParadigm -s <stem> -r <rx for stem> (-e <encoding>) -f <featuresFile> (-o <outputfile>)\n\
   or:\n\
       GenParadigm -h\n"
    sHelpMsg = """
-----------Input format:
 The list of feature structures (in the file given by the -f arg)
 should contain the upper-side names of the affixes, in space-
 delimited form, grouped into lines; the names on each line
 represent a single set of alternatives, usually the alternatives
 for a slot.

 A '^' as the first symbol in a line means that no wordforms
 containing any of the other features in the line will be output
 (useful when outputting a sub-paradigm and a slot is, in essence,
 optional).
 A '0' (zero) at any place in the line means that in addition to
 outputting cells with the forms that contain the other features
 of that line, a cell will be output that contains none of the
 other features in the line (useful when a slot is optional, and
 it is desired to output both the forms containing the affixes
 of that slot, as well as the forms not containing any of those
 affixes).  When the line is negated, the '0' has no effect (but
 may be present, in case some day you decide to reverse the negation).

 Comments may follow a '#' symbol.
 Example:
     #-------Prefixes:
     1SgSubj- 2SgSubj- 3SgSubj- 1PlSubj- 2PlSubj- 3PlSubj-
     1SgObj-  2SgObj-  3SgObj-  1PlObj-  2PlObj-  3PlObj- 0
     #-------Suffixes:
     -Drxnl- 0  #The '0' accounts for the optionality of the -Drxnl
     -PresTns -PastTense -FutTense

-----------Output format:
 The immediate output consist of a series of xfst commands to
 generate the surface forms (if any) corresponding to each
 combination of feature structures, together with the XML tags.
 If run through xfst (with a lexicon loaded and assigned to the
 var 'Lex'), this will give the following XML:
    <?xml version="1.0" encoding="ISO-8859-1"?>
    <!--Generated by GenParadigm.py from Foo.features-->
    <Paradigm stem="foo" selector = "RmBrackets .o. ${[have]}">
       <Cell>
           <Features> ... </Features>
           <Wordforms>
   (wordforms, one per line)
           </Wordforms>
       </Cell>
    </Paradigm>
"""

    #Look for command line args:
    try:
        opts, args = getopt.getopt(sys.argv[1:], 's:r:e:f:o:h')
    except getopt.GetoptError:
        sys.stderr.write(sUsageMsg)
        sys.exit(1)

    #Defaults (or error checking conditions):
    bOutputHelpMsg = False
    sStem = ''
    sSelectStemRX = ''
    sOutFileName = 'stdout'
    sEncoding = 'ISO-8859-1'

    #Non-defaults:
    for o in opts:
        if o[0] == '-s':
            sStem = o[1]
        if o[0] == '-r':
            sSelectStemRX = o[1]
        elif o[0] == '-f':
            sFeaturesFileName = o[1]
        elif o[0] == '-e':
            sEncoding = o[1]
        elif o[0] == '-o':
            sOutFileName = o[1]
        elif o[0] == '-h':
            bOutputHelpMsg = True

    if bOutputHelpMsg:
        #User asked for help msg, give it and exit:
        sys.stderr.write(sUsageMsg)
        sys.stderr.write(sHelpMsg)
        sys.exit(1)

    #If we get here, user did not ask for help msg, so normal run
    # (if everything else is OK).  Make sure we got a stem:
    if sStem == '':
        sys.stderr.write(sUsageMsg)
        sys.exit(1)

    #...and a stem selector:
    if sSelectStemRX == '':
        sys.stderr.write(sUsageMsg)
        sys.exit(1)

    #Open features file:
    if sFeaturesFileName == '':
        sys.stderr.write(sUsageMsg)
    else:
        try:
            FeaturesStream = open(sFeaturesFileName, 'r')
        except IOError:
            sys.stderr.write("GenParadigm: Unable to open features file %s\n" %sFeaturesFileName)
            sys.exit(1)

    #Open output:
    if sOutFileName == 'stdout':
        OutStream = sys.stdout
    else:
        try:
            OutStream = open(sOutFileName, 'w')
        except IOError:
            sys.stderr.write("GenParadigm: Unable to open file %s\n" %sOutFileName)

    return (sStem, sSelectStemRX, FeaturesStream, sEncoding, OutStream)


def ReadlinesWithContinuations(InStream):
    """Read in the lines of InStream, and return them in a list.  This is just
       like the Python readlines(), except that if the last char of a line is
       a backslash, we append the next line (which may also end in a backslash,
       etc.) (replacing the newline with a space char).
       Also, we throw away blank lines and comments."""
    sLines = []
    sCurLine = InStream.readline()
    while sCurLine != '':
        sCurLine = CleanUpLine(sCurLine)
        if sCurLine != '':
            #Not a blank line or only a comment
            while sCurLine != '' and sCurLine[-1] == '\\':
                # (If sCurLine is empty, second test above is safe because first test
                # prevents second one from being tried.)
                #Line ends in a continuation, append next line(s) (with a blank space between).
                sContinuationLine = CleanUpLine(InStream.readline())
                sCurLine = sCurLine[0:-1] + " " + sContinuationLine
            sLines.append(sCurLine)
        #Now get the next line:
        sCurLine = InStream.readline()
    return sLines


def CleanUpLine(sRawLine):
    """Clean up an input line by breaking off any comment, and removing any
       leading or trailing whitespace.  The result may be an empty string."""
    return sRawLine.split('#')[0].strip()


def ParseLine2Features(sLine):
    """Parse the line into a list of zero or more features (represented as strings)
       or negated features (contained in sublists); return the top-level list.
       The line has already been split, and any comments removed."""
    FSublist = re.split('\s+', sLine)
    if FSublist[0] == '^':
        #It's a negated line, so we return a sub-sublist containing the negation
        # of all the remaining features in the line:
        return [FSublist[1:]]
    else:
        #Line is not negated, so just return it as a list:
        return FSublist


def GenerateFeatureStructures(FList, CurrentFStruc):
    "Take the top sublist off FList, choose its next feature.  Then call ourselves \
     recursively on the next sublist of FList until we run out of sublists, \
     at which point we have a complete feature structure.  Once a structure is complete, \
     write it out, then go back to pick the next feature, thereby generating \
     the remaining feature structures recursively, in a depth-first search pattern."
    if FList == []:
        #Recursion has bottomed out, meaning CurrentFStruc is complete; write it out:
        WriteCellFeatureStructure(CurrentFStruc)
    else:
        Head = FList[0]    #Top sublist
        Tail = FList[1:]   #Remaining sublists
        #Choose each feature from the top sublist successively:
        for sFeature in Head:
            if sFeature == '0':
                #A zero means we want a "feature" consisting of the negation
                # of all the other features of the list:
                Head1 = copy.copy(Head)
                Head1.remove('0')   #remove() returns 'None'--the removal is done to Head1
                                    # (so we can't put the .remove() on the line above)
                NewFStruc = CurrentFStruc + [Head1]  #This will be a list of features
                                                     # (to be interpreted as negated)
            else:
                NewFStruc = CurrentFStruc + [sFeature]  #This will be a string (to be
                                                        # interpreted as a single feature)
            #...and do the same for the remaining sublists:
            GenerateFeatureStructures(Tail, NewFStruc)


def WriteCellFeatureStructure(CurrentFStruc):
    "We are given a feature structure, which may contain zeros or sublists. \
     (Sublists represent negated sets of features.) \
     Write the feature structure as an xfst cmd to generate the forms matching \
     that feature structure, bracketed by the echo cmds to generate the \
     descriptive XML tags before and after."
    OutStream.write("echo    <Cell>\necho       <Features> ")
    #We go through the feature structure twice, once to tell xfst to print it out in the XML:
    for F in CurrentFStruc:
        if isinstance(F, str):
            #It must be a feature:
            OutStream.write('%s ' %F)
        else:
            #It must be a list, meaning that we want only the arcs that do not
            # contain the features of the list.  Print a negated bracketed list:
            OutStream.write('^[ ')
            for NegatedF in F:
                if NegatedF != '0':
                    #Ignore zeros
                    OutStream.write('%s ' %NegatedF)
            OutStream.write('] ')
    OutStream.write("</Features>'\n")
    #...and once to generate the wordform(s) matching those features (the format
    # here is almost the same, but not quite):
    OutStream.write("echo    <Wordforms>\n")
    OutStream.write("regex ")
    for F in CurrentFStruc:
        if isinstance(F, str):
            #It must be a feature, meaning we want only the arcs that contain
            # that feature:
            OutStream.write('$"%s" .o. ' %F)
        else:
            #It must be a list, meaning that we want only the arcs that do not
            # contain the features of the list:
            for NegatedF in F:
                if NegatedF != '0':
                    #Ignore zeros
                    OutStream.write('~$"%s" .o. ' %NegatedF)
                #else: shouldn't happen
    #Now make xfst print out the words that are left, which are just those wordforms
    # corresponding to the stem and which match the features list (and don't match
    # the anti-features lists):
    OutStream.write(" LexStemOnly;\nprint lower\npop\n")
    OutStream.write("echo    </Wordforms>\n")
    OutStream.write("echo    </Cell>\n")


#------------------------------main-------------------------------
#Init:
(sStem, sSelectStemRX, FeaturesStream, sEncoding, OutStream) = GetCmdParameters()

#Output xfst cmds, mostly echos.  The first one outputs the XML header:
OutStream.write("echo <?xml version=\"1.0\" encoding=\"%s\"?>\n" %sEncoding)
#...and a comment telling where the output came from:
OutStream.write("echo <!--Generated by GenParadigm.py from %s-->\n"
                %FeaturesStream.name
               )
#...a stylesheet instruction:
OutStream.write('echo <?xml-stylesheet type="text/css" href="Paradigm.xsl"?>\n')
#...the root, whose attrs identify which stem this is the paradigm of, and how that
# stem is to be selected out of all the stems in the xfst lexicon (the latter is
# mainly for debugging purposes, in case you get too many stems--or none):
OutStream.write("echo <Paradigm stem=\"%s\" selector=\"%s\">\n" %(sStem, sSelectStemRX))
#Output an xfst cmd to create XML brackets around each output form.  The first pair of
# backslashes escapes the third backslash, which escapes the quote that we want to put
# around the attr, as opposed to the quote that identifies the extent of the tag in the rule;
# the padding on the left is for indentation for human readers.
OutStream.write('define AddWordformTag [?+ @-> "       <wordform form=\\\""..."\\\"/>"];\n')
#...and the xfst cmd to define a subset of the lexicon which contains
# only the specified stem.  This subset will be used in conjunction with
# the features lists to output only the desired forms:
OutStream.write('define LexStemOnly %s .o. AddWordformTag ;\n' %sSelectStemRX)

#Now generate the feature structures defining the cells of the paradigm:
FList = []
sLines = ReadlinesWithContinuations(FeaturesStream)
FeaturesStream.close()
for sLine in sLines:
    F = ParseLine2Features(sLine)
    if F:
        #Don't append None (returned by ParseLine2Features() if sLine contained no features)
        FList.append(F)
# ...and output the xfst commands that will generate the contents of those cells:
GenerateFeatureStructures(FList, [])

#Finished, tell xfst to generate the close XML tags:
OutStream.write("echo </Paradigm>\n")
OutStream.close()
